;(function (name, definition) {
  'use strict';
  if (typeof define == 'function') {
    define(definition)
  } else if (typeof module != 'undefined' && module.exports) {
    module.exports = definition()
  } else {
    var Module = definition(), global = this, old = global[name]
    Module.noConflict = function () {
      global[name] = old
      return Module
    }
    global[name] = Module
  }
}).call(this, 'fixMyJS', function () {
  var fixMyJS = require('./legacy')

  var esprima = require('esprima')
  var escodegen = require('escodegen')
  var fu = require('fu')
  var rules = require('./rules')

  var validRules = {
    indentpref: ['spaces', 'tabs'],
    quotmark: ['single', 'double']
  }

  function getRules(has) {
    var getRule = function (x) { return x[1] }
    var falseRule = function (x) { return has(x[0], false) }
    var trueRule = fu.comp(fu.not, falseRule)
    return fu.concat(
      rules.builtin,
      fu.map(getRule, fu.filter(falseRule, fu.intoArray(rules.aretrue))),
      fu.map(getRule, fu.filter(trueRule, fu.intoArray(rules.arefalse)))
    )
  }

  function validateRules(rules) {
    if (rules.camelcase && rules.snakecase) {
      throw new Error('Cannot contain both camelcase and snakecase options')
    }

    fu.map(fu.apply(function (ruleName, possibleValues) {
      if (rules.hasOwnProperty(ruleName) &&
          !fu.elem(rules[ruleName], possibleValues)) {
        throw new Error('Invalid type for ' + ruleName)
      }
    }), fu.intoArray(validRules))

    return rules
  }

  function traverse(o, f, p) {
    var k, r = Array.isArray(o) ? [] : {}
    var self = function (x) { return traverse(x, f, p) }
    for (k in o) {
      if (toString.call(o[k]) == '[object Object]') {
        r[k] = traverse(o[k], f, o)
      } else if (Array.isArray(o[k])) {
        r[k] = fu.concatMap(self, o[k])
      } else {
        r[k] = o[k]
      }
    }
    return f(r, p)
  }

  function createIndent(n, indent) {
    return Array(Number(n) + 1).join(indent == 'spaces' ? ' ' : '\t')
  }

  function genHas(obj) {
    return function (key, fallback) {
      return obj.hasOwnProperty(key) && obj[key] !== undefined
        ? obj[key]
        : fallback
    }
  }

  fixMyJS.fix = function (code, config) {
    var ast = esprima.parse(code, { comment: true })
    var has = genHas(validateRules(config))
    var rules = getRules(has)
    var options = {
      format: {
        indent: {
          style: createIndent(has('indent', 2), has('indentpref', 'spaces')),
          base: 0
        },
        json: false,
        renumber: false,
        quotes: has('quotmark', 'single'),
        escapeless: false,
        parentheses: true,
        semicolons: !has('asi', false)
      },
      parse: null,
      comment: true
    }
    var modifiedTree = traverse(ast, function (node, parent) {
      return fu.foldl(function (node, f) {
        return f.hasOwnProperty(node.type)
          ? f[node.type](node, parent)
          : node
      }, rules, node)
    })
    return escodegen.generate(modifiedTree, options)
  }

  fixMyJS.version = require('../package.json').version

  return fixMyJS
})
